#include "clb.h"
#include <stdio.h>

#define CALLBACK_NONE   0
#define CALLBACK_SIGNAL 1
#define CALLBACK_ALIAS  2

#define STREAM_STDOUT 0
#define STREAM_STDERR 1

#define WM_DISPATCHBUFFER (WM_USER + 1)

static DWORD mainThreadId;
static HWND mWnd;

typedef struct ExecPipe {
	int		callbackType;
	char   *callbackName;
	int		streamType;
	HANDLE	streamRead;
	HANDLE  streamWrite;
} ExecPipe;

void PushData(ExecPipe *pipe, const char *buffer) {
	char *prefix, *streamTypeName, *outbuf = malloc(MIRC_MAXSTRING);

	switch(pipe->callbackType) { 
		case CALLBACK_SIGNAL: prefix = ".signal -n "; break;
		default:
			prefix = "";
	}

	switch(pipe->streamType) {
		case STREAM_STDOUT: streamTypeName = "stdout"; break;
		case STREAM_STDERR: streamTypeName = "stderr"; break;
		default:
			streamTypeName = "unknown";
	}

	_snprintf(outbuf, MIRC_MAXSTRING, "/%s%s %s %s", prefix, pipe->callbackName, streamTypeName, buffer);

	PostThreadMessage(mainThreadId, WM_DISPATCHBUFFER, (WPARAM)pipe, (LPARAM)outbuf);
}

DWORD WINAPI ReadPipe(ExecPipe *pipe) {
	int readBytes, totalBytes = 0;
	int callbackNameLength = pipe->callbackName ? (int)strlen(pipe->callbackName) : 0;
	char buffer[MIRC_MAXSTRING];

	while (ReadFile(pipe->streamRead, buffer+totalBytes, 1, &readBytes, NULL)) {
		if (readBytes == 0) break;

		totalBytes += readBytes;

		if (buffer[totalBytes-1] == '\n' ||		/* Newline or max string */
			totalBytes >= MIRC_MAXSTRING - callbackNameLength - 
			(pipe->callbackType == CALLBACK_SIGNAL ? 38 : 20)) {
			/* On a newline just ignore the \r and \n characters */
			if (buffer[totalBytes-1] == '\n') totalBytes--;
			if (buffer[totalBytes-1] == '\r') totalBytes--;

			buffer[totalBytes] = 0;
			PushData(pipe, buffer);

			/* Reset buffer */
			totalBytes = 0; 
		}
    }

	if (totalBytes > 0) { /* Flush the rest of the buffer */
		buffer[totalBytes+1] = 0;
		PushData(pipe, buffer);
	}

	return 0;
}

BOOL CALLBACK TerminateAppEnum(HWND hwnd, LPARAM lParam) {
	DWORD dwID;
	
	GetWindowThreadProcessId(hwnd, &dwID);

	if (dwID == (DWORD)lParam) {
		PostMessage(hwnd, WM_CLOSE, 0, 0);
	}
	return TRUE;
}

#define NUMTHREADS 2
BOOL ExecuteCommand(const char *command, int callbackType, char *callbackName) {
	ExecPipe exc[NUMTHREADS] = {0};
	HANDLE threads[NUMTHREADS] = {NULL};
	DWORD threadId[NUMTHREADS] = {0};
	MSG msg;
	int activeThreads, numMsgs, code;
	BOOL broken = FALSE;

	STARTUPINFOA StartupInfo;
	PROCESS_INFORMATION ProcessInfo;
    SECURITY_ATTRIBUTES secAtrib;

	int i;
	int commandlen = (int)strlen(command);
    char *cmd = getenv("COMSPEC") != NULL ? getenv("COMSPEC") : "COMMAND.COM";
    char *args = malloc(commandlen + 5);

	/* Set arguments */
	strcpy(args, " /c ");
	strncat(args, command, commandlen); 

    /* Create STDOUT pipe */
    secAtrib.nLength = sizeof(SECURITY_ATTRIBUTES);
    secAtrib.bInheritHandle = TRUE;
    secAtrib.lpSecurityDescriptor = NULL;
    
	/* Initialize ExecPipes and the pipes themselves */
	if (callbackType != CALLBACK_NONE) {
		for (i = 0; i < NUMTHREADS; i++) {
			exc[i].callbackName = callbackName;
			exc[i].callbackType = callbackType;
			exc[i].streamType = i;

			CreatePipe(&exc[i].streamRead, &exc[i].streamWrite, &secAtrib, 0);
			SetHandleInformation(exc[i].streamRead, HANDLE_FLAG_INHERIT, 0);
		}
	}

    /* Set Startup information for process */
	memset(&StartupInfo, 0, sizeof(StartupInfo));
	StartupInfo.cb = sizeof(STARTUPINFO);
    StartupInfo.dwFlags = STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;
	StartupInfo.hStdInput = NULL;
	if (callbackType != CALLBACK_NONE) {
	    StartupInfo.hStdOutput = exc[STREAM_STDOUT].streamWrite;
		StartupInfo.hStdError = exc[STREAM_STDERR].streamWrite;
	}
    StartupInfo.wShowWindow = SW_HIDE;

    if (CreateProcessA(cmd, args, NULL, NULL, TRUE, CREATE_NEW_CONSOLE, NULL, NULL, &StartupInfo, &ProcessInfo)) {

		if (callbackType != CALLBACK_NONE) {
			/* Record this thread id for the pipes to post messages to */
			mainThreadId = GetCurrentThreadId();

			for (i = 0; i < NUMTHREADS; i++) {
				CloseHandle(exc[i].streamWrite); /* Write stream must be closed before reading 
												  * or ReadFile will block. */

				threads[i] = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ReadPipe, &exc[i], 0, &threadId[i]);
			}

			/* Start the message loop awaiting threads to speak up */
			while (1) {
				/* Let's try to catch ctrl-break */
				if (PeekMessage(&msg, NULL, WM_KEYDOWN, WM_KEYUP, PM_REMOVE|PM_NOYIELD)) {
					if (msg.wParam == 3) { /* Check for ctrl-break */
						broken = TRUE;
						break;	
					}
				}
				
				if (numMsgs = PeekMessage(&msg, (HWND)-1, WM_USER, WM_APP, PM_REMOVE|PM_NOYIELD)) {
					if (msg.message == WM_DISPATCHBUFFER) { /* Handle buffer output */		
						CLBCommand((char *)msg.lParam, 0);
						free((void *)msg.lParam);
					}
				}

				/* Count how many threads we have left */
				activeThreads = 0;
				for (i = 0; i < NUMTHREADS; i++) {
					GetExitCodeThread(threads[i], &code);
					if (code == STILL_ACTIVE) {
						activeThreads++;
					}
				}

				/* Check if threads are dead and no messages are on the queue */
				if (!activeThreads && !numMsgs) { /* We're done */
					break;
				}
			}
		}

		if (!broken) {
			/* The process should have ended by now */
			WaitForSingleObject(ProcessInfo.hProcess, INFINITE);
		}
		else { /* Called ctrl-break, dirty cleanup */
			EnumWindows((WNDENUMPROC)TerminateAppEnum, (LPARAM)ProcessInfo.dwProcessId);

			if (WaitForSingleObject(ProcessInfo.hProcess, 1000) != WAIT_OBJECT_0) {
				/* It did not want to die, this probably won't work */
				TerminateProcess(ProcessInfo.hProcess, -1);
			}
			
			for (i = 0; i < NUMTHREADS; i++) {
				TerminateThread(threads[i], 1);
			}
		}
	}

#ifndef _DEBUG
	for (i = 0; i < NUMTHREADS; i++) {
		/* Close all open handles */
		CloseHandle(exc[i].streamRead);
		CloseHandle(exc[i].streamWrite);
		CloseHandle(threads[i]);
	}
#endif

	CloseHandle(ProcessInfo.hThread);
	CloseHandle(ProcessInfo.hProcess);

	free(args);

	return !broken;
}

int __declspec(dllexport) __stdcall execS(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show, BOOL nopause) {
	char *command, *signal;

	if (!CLBInitialize(&data)) return 2;

	signal = data;
	command = strchr(data, ' ');
	if (command == NULL) {
		strcpy(data, "/echo $color(info) * /exec: missing signal name");
		return 2;
	}
	*command++ = 0; /* Get character after space */

	if (!ExecuteCommand(command, CALLBACK_SIGNAL, signal)) {
		CLB_HALT(data);
	}

	return 1;
}

int __declspec(dllexport) __stdcall execA(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show, BOOL nopause) {
	char *command, *alias;

	if (!CLBInitialize(&data)) return 2;

	alias = data;
	command = strchr(data, ' ');
	if (command == NULL) {
		strcpy(data, "/echo $color(info) * /exec: missing alias callback");
		return 2;
	}
	*command++ = 0; /* Get character after space */

	if (!ExecuteCommand(command, CALLBACK_ALIAS, alias)) {
		CLB_HALT(data);
	}

	return 1;
}

int __declspec(dllexport) __stdcall exec(HWND mWnd, HWND aWnd, char *data, char *parms, BOOL show, BOOL nopause) {
	if (!CLBInitialize(&data)) return 2;

	if (!ExecuteCommand(data, CALLBACK_NONE, NULL)) {
		CLB_HALT(data);
	}

	return 1;
}

void __declspec(dllexport) __stdcall LoadDll(LOADINFO *t) {
	mWnd = t->mHwnd;
	CLBStartup(t, NULL);
}

int  __declspec(dllexport) __stdcall UnloadDll(int mTimeout) {
    if (mTimeout == 0) { /* user called /dll -u */
		CLBShutdown();
    }
	return 1;
}